From 5df6ca67588e2e9ad3527eb9030634599f6ee9c1 Mon Sep 17 00:00:00 2001
From: Shugo Maeda <shugo@ruby-lang.org>
Date: Thu, 29 Jul 2010 14:12:25 +0900
Subject: [PATCH 12/51] binding(1) returns the binding of the caller.

---
 eval.c                |    2 +-
 include/ruby/intern.h |    2 +-
 proc.c                |   23 +++++++++++++++++++----
 thread.c              |    2 +-
 vm.c                  |    2 +-
 5 files changed, 23 insertions(+), 8 deletions(-)

diff --git a/eval.c b/eval.c
index 33dc10a..89a9862 100644
--- a/eval.c
+++ b/eval.c
@@ -20,7 +20,7 @@
 #define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
 
 VALUE proc_invoke(VALUE, VALUE, VALUE, VALUE);
-VALUE rb_binding_new(void);
+VALUE rb_binding_new(int);
 NORETURN(void rb_raise_jump(VALUE));
 
 ID rb_frame_callee(void);
diff --git a/include/ruby/intern.h b/include/ruby/intern.h
index 7cc75b2..100c026 100644
--- a/include/ruby/intern.h
+++ b/include/ruby/intern.h
@@ -339,7 +339,7 @@ VALUE rb_proc_call(VALUE, VALUE);
 VALUE rb_proc_call_with_block(VALUE, int argc, VALUE *argv, VALUE);
 int rb_proc_arity(VALUE);
 VALUE rb_proc_lambda_p(VALUE);
-VALUE rb_binding_new(void);
+VALUE rb_binding_new(int);
 VALUE rb_obj_method(VALUE, VALUE);
 VALUE rb_method_call(int, VALUE*, VALUE);
 int rb_mod_method_arity(VALUE, ID);
diff --git a/proc.c b/proc.c
index 87cf3c1..de2ed3d 100644
--- a/proc.c
+++ b/proc.c
@@ -311,13 +311,21 @@ binding_clone(VALUE self)
 }
 
 VALUE
-rb_binding_new(void)
+rb_binding_new(int level)
 {
     rb_thread_t *th = GET_THREAD();
     rb_control_frame_t *cfp = rb_vm_get_ruby_level_next_cfp(th, th->cfp);
     VALUE bindval = binding_alloc(rb_cBinding);
     rb_binding_t *bind;
+    int i;
 
+    if (level < 0) {
+	rb_raise(rb_eArgError, "level should not be negative: %d", level);
+    }
+    for (i = 0; i < level && cfp; i++) {
+	cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp);
+	cfp = rb_vm_get_ruby_level_next_cfp(th, cfp);
+    }
     if (cfp == 0) {
 	rb_raise(rb_eRuntimeError, "Can't create Binding Object on top of Fiber.");
     }
@@ -346,9 +354,16 @@ rb_binding_new(void)
  */
 
 static VALUE
-rb_f_binding(VALUE self)
+rb_f_binding(int argc, VALUE *argv, VALUE self)
 {
-    return rb_binding_new();
+    VALUE vlevel;
+    int level = 0;
+
+    rb_scan_args(argc, argv, "01", &vlevel);
+    if (!NIL_P(vlevel)) {
+	level = NUM2INT(vlevel);
+    }
+    return rb_binding_new(level);
 }
 
 /*
@@ -2238,6 +2253,6 @@ Init_Binding(void)
     rb_define_method(rb_cBinding, "clone", binding_clone, 0);
     rb_define_method(rb_cBinding, "dup", binding_dup, 0);
     rb_define_method(rb_cBinding, "eval", bind_eval, -1);
-    rb_define_global_function("binding", rb_f_binding, 0);
+    rb_define_global_function("binding", rb_f_binding, -1);
 }
 
diff --git a/thread.c b/thread.c
index c28ea9c..9545eb8 100644
--- a/thread.c
+++ b/thread.c
@@ -4187,7 +4187,7 @@ call_trace_proc(VALUE args, int tracing)
     argv[1] = filename;
     argv[2] = INT2FIX(line);
     argv[3] = id ? ID2SYM(id) : Qnil;
-    argv[4] = (p->self && srcfile) ? rb_binding_new() : Qnil;
+    argv[4] = (p->self && srcfile) ? rb_binding_new(0) : Qnil;
     argv[5] = klass ? klass : Qnil;
 
     return rb_proc_call_with_block(p->proc, 6, argv, Qnil);
diff --git a/vm.c b/vm.c
index b9bd45c..7d81d66 100644
--- a/vm.c
+++ b/vm.c
@@ -2134,7 +2134,7 @@ Init_VM(void)
 	th->cfp->pc = iseq->iseq_encoded;
 	th->cfp->self = th->top_self;
 
-	rb_define_global_const("TOPLEVEL_BINDING", rb_binding_new());
+	rb_define_global_const("TOPLEVEL_BINDING", rb_binding_new(0));
     }
     vm_init_redefined_flag();
 }
-- 
1.7.0.4

